@@ -12,6 +12,10 @@ from paste.httpexceptions import HTTPExceptionHandler
 import paste.request
 from paste.request import construct_url
 from paste.util.import_string import eval_import
+import Cookie
+import md5
+import socket
+import struct
 
 
 class AuthFormHandler(object):
@@ -24,6 +28,9 @@ class AuthFormHandler(object):
     def __init__(self, app, app_conf, authform_conf):
         self.app = app
         self.loginurl = authform_conf.get('loginurl', '/login')
+        self.loginurl_func = authform_conf.get('loginurlfunc', None)
+        if isinstance(self.loginurl_func, str):
+            self.loginurl_func = eval_import(self.loginurl_func)
         self.session_middleware = authform_conf.get('session_middleware', 'beaker.session')
 
         self.process_url = authform_conf.get('processurl', '/process')
@@ -37,6 +44,38 @@ class AuthFormHandler(object):
         if isinstance(self.is_authenticated, str):
             self.is_authenticated = eval_import(self.is_authenticated)
 
+        self.rememberme = authform_conf.get('rememberme', True)
+        self.rememberme_param = authform_conf.get('rememberme_param', 'rememberme')
+        self.cookie_name = authform_conf.get('cookiename', 'authformcookie')
+        self.cookie_expires = authform_conf.get('cookieexpires', 3600.0)
+        self.ip_mask = authform_conf.get('ipmask', '255.255.255.0')
+        self.cookie_secret = authform_conf.get('cookiesecret', 'secret')
+        # Use agent string if you are sure that your site uses single agent.
+        # E.g. SWFUpload generates different user agent string.
+        self.cookie_use_agent = bool(authform_conf.get('cookieuseagent', 'False'))
+
+    def _generate_cookie_value(self, environ):
+        address = environ.get('REMOTE_ADDR','0.0.0.0')
+        mask = struct.unpack('>L', socket.inet_aton(self.ip_mask))[0]
+        address = struct.unpack('>L', socket.inet_aton(address))[0]
+
+        user_string = socket.inet_ntoa(struct.pack('>L', address & mask))
+        if self.cookie_use_agent:
+            user_string += environ.get('HTTP_USER_AGENT', '')
+        user_string += self.cookie_secret
+        return md5.md5(user_string).hexdigest()
+
+    def _check_authformcookie(self, environ):
+        cookie_value = ''
+        cookies = environ.get('HTTP_COOKIE', '')
+        idx = cookies.find(self.cookie_name + '=')
+        if idx != -1:
+            cookie_value = cookies[idx + len(self.cookie_name)+1:]
+            cookie_end = cookie_value.find(';')
+            if cookie_end != -1:
+                cookie_value = cookie_value[:cookie_end]
+        return self._generate_cookie_value(environ) == cookie_value
+
     def __call__(self, environ, start_response):
         status_ = []
         headers_ = []
@@ -54,6 +93,7 @@ class AuthFormHandler(object):
 
         if environ.get('PATH_INFO') == self.process_url:
             response = self.process(environ, start_response)
+            session.save()
             return response
         elif environ.get('PATH_INFO') == self.logout_url:
             if session.has_key('REMOTE_USER'):
@@ -67,15 +107,23 @@ class AuthFormHandler(object):
             headers_.append(headers)
             exc_info_.append(exc_info)
 
-        if session.get('REMOTE_USER'):
+        if self._check_authformcookie(environ) and session.get('REMOTE_USER'):
+            environ['AUTH_TYPE'] = 'form'
             environ['REMOTE_USER'] = session['REMOTE_USER']
             environ['x-wsgiorg.user_data'] = session.get('x-wsgiorg.user_data', {})
+        elif session.has_key('x-wsgiorg.auth_error'):
+            environ['x-wsgiorg.auth_error'] = session.pop('x-wsgiorg.auth_error')
+            session.save()
 
         result = self.app(environ, fake_start_response)
         if status_[0][:3] == '401':
             session['referer'] = construct_url(environ)
             session.save()
-            start_response('301 Redirect', [('Content-type', 'text/html'), ('Location', self.loginurl)])
+            if self.loginurl_func is not None:
+                loginurl = self.loginurl_func(environ)
+            else:
+                loginurl = self.loginurl
+            start_response('301 Redirect', [('Content-type', 'text/html'), ('Location', loginurl)])
             return []
         else:
             start_response(status_[0], headers_[0], exc_info_[0])
@@ -87,20 +135,37 @@ class AuthFormHandler(object):
         params = dict(paste.request.parse_formvars(environ))
 
         session = environ[self.session_middleware]
-        authenticated, data = self.is_authenticated(environ, params['username'], params['password'])
+        authenticated, data = self.is_authenticated(environ, params.get('username', None), params.get('password', None))
         if authenticated:
             session['REMOTE_USER'] = params['username']
             if data:
                 session['x-wsgiorg.user_data'] = data
-            if 'referer' in session:
+            if 'redirect' in params:
+                redirect_url = params['redirect']
+            elif 'referer' in session:
                 redirect_url = session.pop('referer')
             else:
                 redirect_url = self.baseurl + self.loggedin_url
             session.save()
-            start_response('301 Redirect', [('Content-type', 'text/html'), ('Location', redirect_url)])
+            cookie = Cookie.SimpleCookie()
+            cookie[self.cookie_name] = self._generate_cookie_value(environ)
+            cookie[self.cookie_name]['path'] = '/'
+            if not self.rememberme or self.rememberme_param in params:
+                cookie[self.cookie_name]['expires'] = Cookie._getdate(self.cookie_expires)
+
+            start_response('301 Redirect', [('Content-type', 'text/html'),
+                ('Location', redirect_url),
+                ('Set-Cookie', cookie.output()[len('Set-Cookie: '):])])
             return []
+        else:
+            session['x-wsgiorg.auth_error'] = data
+
+        if self.loginurl_func is not None:
+            loginurl = self.loginurl_func(environ)
+        else:
+            loginurl = self.loginurl
+        start_response('301 Redirect', [('Content-type', 'text/html'), ('Location', loginurl)])
 
-        start_response('301 Redirect', [('Content-type', 'text/html'), ('Location', self.loginurl)])
         return []
 
 
